
/*
 * Zap user and group IDs.  Part of the passwd-file-UID-matching
 * program.
 */
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/quota.h>
#include <sys/inode.h>
#include <sys/fs.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fstab.h>
#include <netinet/in.h>
#include "mf.h"

struct	mapfent *mf;		/* the map */
int	nmf;			/* and its size */
int	mtmp;		
int	olduid, newuid;

#define	MAPUID(uid) ((uid) >= nmf ? -1 : mf[uid].mf_uid)
#define	MAPGID(gid) ((gid) >= nmf ? -1 : mf[gid].mf_gid)

main(argc, argv)
	register int argc;
	register char **argv;
{
	register int code = 0, i;

	if (argc <= 3)
		err(1, "usage: %s olduid newuid quotafile ", *argv);
	olduid = atoi(*++argv);
	newuid = atoi(*++argv);
	checkqfile(*++argv);
	readmap();
}

int	qfd, nfd;
char	outf[100];

checkqfile(str)
char	*str;
{
	printf("quotafile is %s\n", str);
	if ((qfd = open(str, O_RDONLY)) < 0 )
		err(1, "cannot open quota file %s", str);
	strcpy(outf, str);
	strcat(outf, ".new");
	/*if ((nfd = open(outf, O_CREAT|O_WRONLY)) < 0 )*/
	if ((nfd = open(outf, O_RDONLY)) < 0 )
		err(1, "cannot open new quota file %s", outf);
}

readmap()
{
	register struct mapfent *m;
	int fd;
	short magic;
	struct stat st;
	char fn[100];
	struct  dqblk odqblk, ndqblk;

	(void) strcpy(fn, "/map");
	fd = open(fn, 0);
	if (fd < 0) {
		fn[0] = '/';
		(void) gethostname(fn + 1, sizeof fn - 6);
		(void) strcat(fn, ".map");
		fd = open(fn, 0);
		if (fd < 0) {
			printf ("error in open\n");
			fflush();
			err(1, "cannot open /map or %s: %m", fn);
		}
	}
	if (fstat(fd, &st)) {
		printf ("error in fstat\n");
		fflush();
		err(1, "fstat: %m");
	}
	st.st_size -= sizeof (magic);
	if (st.st_size < 0 || (st.st_size % sizeof (*mf)) != 0 ||
	    read(fd, (char *) &magic, sizeof (magic)) != sizeof (magic) ||
	    ntohs(magic) != MAPMAGIC) {
		printf("map file error - magic is %x\n", (int)magic);
		fflush();
		err(1, "%s is not a proper map file", fn);
	}
	if ((mf = (struct mapfent *) malloc((unsigned) st.st_size)) == NULL) {
		printf("malloc error\n");
		fflush();
		err(1, "cannot allocate %d map bytes: %m", st.st_size);
	}
	if (read(fd, (char *) mf, st.st_size) != st.st_size) {
		printf("read bomb\n");
		fflush();
		err(1, "map read: %m");
	}
	nmf = st.st_size / sizeof (*mf);
	printf("number of entries is %d\n", nmf);
	printf("sizeof quota struct is %d\n", sizeof odqblk);
	for (m = &mf[nmf]; --m >= mf;) {
		m->mf_uid = ntohl(m->mf_uid);
		m->mf_gid = ntohl(m->mf_gid);
	}
	for (mtmp = nmf, m = &mf[nmf]; --m >= mf;) {
		--mtmp;
		if(m->mf_uid == -1) continue;
		if(mtmp !=  olduid) continue;
		printf("map entry %d, new uid is %d\n", mtmp, m->mf_uid);
		printf("will seek %d bytes in quota file\n", olduid * sizeof odqblk);
		lseek(qfd, (long)(olduid * sizeof (odqblk)), L_SET);
		if(read(qfd, &odqblk, sizeof odqblk) != sizeof odqblk){
			perror("read");
			err(1, "bad read from quota file");
		}
		printf("will seek %d bytes in quota file\n", m->mf_uid * sizeof odqblk);
		lseek(nfd, (long)(m->mf_uid * sizeof (odqblk)), L_SET);
		if(read(nfd, &ndqblk, sizeof odqblk) != sizeof odqblk)
			err(1, "bad read to old quota file");
		printf("old limits were %ld and %ld\n", odqblk.dqb_bhardlimit,
				odqblk.dqb_bsoftlimit);
		printf("new limits are %ld and %ld\n", ndqblk.dqb_bhardlimit,
				ndqblk.dqb_bsoftlimit);
	}

}

