#ifndef lint
static char rcsid[] = "$Header: map.c,v 1.1 86/10/04 03:33:14 chris Exp $";
#endif

#include <stdio.h>
#include <sys/types.h>
#include "alloc.h"
#include "map.h"

char	*malloc();

/*
 * Allocate a map entry.  The new entry is zero filled.
 */
struct mapent *
mealloc(map)
	register struct map *map;
{
	register caddr_t cp;
	register int n;
	register u_int siz = map->m_size;
#define cptome(x) ((struct mapent *) (x))

	if ((cp = (caddr_t) map->m_free) == NULL) {
		n = ntoalloc(siz);
		if ((cp = malloc(n * siz)) == NULL)
			err(1, "mealloc: out of memory: %m");
		--n;
		cp += n * siz;
		cptome(cp)->me_link = NULL;
		while (--n >= 0) {
			cptome(cp - siz)->me_link = cptome(cp);
			cp -= siz;
		}
	}
	map->m_free = cptome(cp)->me_link;
	bzero(cp, siz);
	return (cptome(cp));
#undef cptome
}

/*
 * Clear an entire map.
 */
void
mapfree(map)
	register struct map *map;
{
	register struct mapent **mep, *me;
	register int n;

	for (n = 0, mep = map->m_map; n < MHSZ; n++, mep++) {
		if ((me = *mep) != NULL) {
			for (;;) {
				if (me->me_name)
					free(me->me_name);
				if (me->me_link == NULL)
					break;
				me = me->me_link;
			}
			me->me_link = map->m_free;
			map->m_free = *mep;
			*mep = 0;
		}
	}
}

/*
 * Find or create a map entry for the given name or key.
 * If created rather than found, *isnew will be set.
 * If isnew is NULL, return NULL instead of creating a new entry.
 */
struct mapent *
getme(map, name, key, isnew)
	struct map *map;
	char *name;
	int key;
	register int *isnew;
{
	register int h;
	register struct mapent **mep, *me;
	register char *p;

	if (isnew != NULL)
		*isnew = 0;

	if ((p = name) == NULL) {
		h = key;
		for (mep = &map->m_map[(unsigned) h % MHSZ];
		    (me = *mep) != NULL; mep = &me->me_link)
			if (me->me_hash == h)
				return (me);
	} else {
		h = 0;
		while (*p)
			h = (h << 5) - h + *p++;
		p = name;
		for (mep = &map->m_map[(unsigned) h % MHSZ];
		    (me = *mep) != NULL; mep = &me->me_link)
			if (me->me_hash == h /* && *p == *me->me_name */ &&
			    strcmp(p, me->me_name) == 0)
				return (me);
	}

	if (isnew == NULL)
		return (NULL);
	*mep = me = mealloc(map);
	me->me_hash = h;
	me->me_name = savestr(p);
	*isnew = 1;
	return (me);
}

/*
 * Apply a function to all members of a map.
 */
mapapply(map, f)
	register struct map *map;
	register int (*f)();
{
	register struct mapent *me;
	register int n;

	for (n = 0; n < MHSZ; n++)
		for (me = map->m_map[n]; me != NULL; me = me->me_link)
			(*f)(me);
}
