/*
 * Maps map from keys (names or integers) to some arbitrary unit of
 * information through a map table.  Names are hashed and the code
 * is used as an index into a chain head.  Integer keys are used directly
 * as a hash.
 */
#define	MHSZ	1009			/* hash table size */

/*
 * Map entries have a hash code, an optional name, and a chain link.
 * This is followed by the data, which is of indeterminate size (but
 * is always at least one integer's worth).
 */
struct mapent {
	int	me_hash;		/* key if int; always hash value */
	char	*me_name;		/* key if name */
	struct	mapent *me_link;	/* chain */
	int	me_data;		/* data, unknown size */
};

struct map {
	u_int	m_size;			/* size of an entry's data */
	struct	mapent *m_free;		/* free list */
	struct	mapent *m_map[MHSZ];	/* hash chain heads */
};

/*
 * Convert a map entry to a data pointer, or a data pointer to
 * a map entry.
 */
#define	metod(m, ty)	((ty) &(m)->me_data)
#define dtome(x)	((struct mapent *) (((int *) (x) + 1) - 1))

/*
 * Return the size of a proper map entry for an object of the given type.
 */
#define	mesize(ty)	(sizeof (ty) + sizeof (struct mapent) - sizeof (int))

/*
 * Function types.
 */
#ifdef ANSI
struct	mapent *mealloc(struct map *map);
void	mapfree(struct map *map);
struct	mapent *getme(struct map *map, char *name, int key, int *isnew);
#else
struct	mapent *mealloc();
void	mapfree();
struct	mapent *getme();
#endif
