#ifndef lint
static char rcsid[] = "$Header: mp.c,v 1.2 86/10/05 02:34:05 chris Exp $";
#endif

/*
 * mp - make new passwd files given a bunch of inputs.
 * Also creates mapping tables.
 */

#include <sys/param.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include "alloc.h"
#include "map.h"
#include "mygrp.h"
#include "mypwd.h"
#include "splice.h"
#include "subr.h"

/*
 * Per-user global information.
 */
struct guinfo {
	int	gu_uid;			/* new UID */
	int	gu_gid;			/* new GID */
};

/*
 * Global (new) group information.
 */
struct groupinfo {
	int	gi_gid;			/* new gid */
	int	gi_nu;			/* number of users in gu */
	int	gi_nf;			/* number of free slots in gu */
	char	**gi_gu;		/* list of users */
};

/*
 * Maps.
 */
struct map gumap = { mesize(struct guinfo) };	/* global user map */
struct map groupmap = { mesize(char *) };	/* group id to name */
struct map gimap = { mesize(struct groupinfo) };/* master group name to id */

/*
 * Global variables.
 */
char	*machine;		/* current host name */
int	nextuid = 256;		/* uids 0-255 `reserved' */
int	nextgid = 128;		/* gids 0-127 `reserved' */

char	*index(), *malloc(), *realloc();

/*
 * `Fast' test for string equality.
 */
#define	EQ(a, b)	(*(a) == *(b) && strcmp(a, b) == 0)

main(argc, argv)
	register int argc;
	register char **argv;
{
	register int i;
	int nmasters = 0;
	extern char *optarg;
	extern int optind;

	while ((i = getopt(argc, argv, "g:m:u:")) != EOF) {
		switch (i) {

		case 'g':
			nextgid = atoi(optarg);
			break;

		case 'u':
			nextuid = atoi(optarg);
			break;

		case 'm':
			domaster(optarg);
			nmasters++;
			break;

		case '?':
usage:
			err(1, "\
usage: %s [-u uid] [-g gid] [-m master] host ...",
				*argv);
			/* NOTREACHED */

		default:
			err(1, "bug! case %c?", i);
			/* NOTREACHED */
		}
	}

	if (optind >= argc)
		goto usage;

	if (nmasters == 0)
		domaster("base");

	for (i = optind; i < argc; i++)
		doslave(argv[i]);
	gengrp();
	exit(0);
}

/*
 * Convert a group ID to a group name.
 */
char *
gidtogroup(pw)
	struct passwd *pw;
{
	struct mapent *me;

	me = getme(&groupmap, (char *) NULL, pw->pw_gid, (int *) NULL);
	if (me == NULL) {
		err(0, "gid %d, user %s@%s: cannot find group name",
			pw->pw_gid, pw->pw_name, machine);
		return ("dullgrp");
	}
	return (*metod(me, char **));
}

/*
 * Initialise a newly created groupinfo entry.
 */
giinit(gi, gid)
	register struct groupinfo *gi;
	int gid;
{

	gi->gi_gid = gid < 0 ? nextgid++ : gid;
	gi->gi_nu = 0;
	gi->gi_nf = 1;
	gi->gi_gu = (char **) malloc(sizeof (char *));
	if (gi->gi_gu == NULL)
		err(1, "giinit: out of memory: %m");
}

/*
 * Return the master group ID for the given group.
 */
int
mastergid(name)
	char *name;
{
	struct groupinfo *gi;
	int new;

	gi = metod(getme(&gimap, name, 0, &new), struct groupinfo *);
	if (new)
		giinit(gi, -1);
	return (gi->gi_gid);
}

/*
 * Add a group for a user.
 */
useringroup(user, group)
	char *user;
	register char *group;
{
	register struct groupinfo *gi;
	register int n;
	int new;

	gi = metod(getme(&gimap, group, 0, &new), struct groupinfo *);
	if (new)
		giinit(gi, -1);

	/* linear search: inefficient? */
	for (n = 0; n < gi->gi_nu; n++)
		if (EQ(user, gi->gi_gu[n]))
			return;

	if (gi->gi_nf <= 0) {
		gi->gi_gu = (char **) realloc((char *) gi->gi_gu,
			(unsigned) (gi->gi_nu * 2 * sizeof (char *)));
		if (gi->gi_gu == NULL)
			err(1, "useringroup: out of memory: %m");
		gi->gi_nf = gi->gi_nu;
	}
	gi->gi_gu[gi->gi_nu++] = poolstr(user);
	gi->gi_nf--;
}

/*
 * Enter a (user, machine) tuple.
 * In case this is the first occurrence of the given user
 * name, we need a complete password entry anyway, so instead
 * of name, uid, and gid we take the whole passwd structure.
 */
enteruser(who, uid)
	register struct passwd *who;
	int uid;
{
	register struct guinfo *gu;
	int created;

	gu = metod(getme(&gumap, who->pw_name, 0, &created), struct guinfo *);
	if (created) {
		gu->gu_uid = uid < 0 ? nextuid++ : uid;
		gu->gu_gid = mastergid(gidtogroup(who));
	}
	useringroup(who->pw_name, gidtogroup(who));
#ifdef notdef			/* done later */
	putmapelt(pw->pw_uid, gu->gu_uid);
#endif
}

/*
 * Read a group file and add to group map(s).  We always add to the
 * local map, which maps from IDs to names; if mastermap is set, we
 * also add to the global map, which maps from names to IDs.  When
 * doing a regular (non-master) map, we also add each user name to
 * its appropriate group.
 */
dogroupmap(file, mastermap)
	char *file;
	int mastermap;
{
	register struct mapent *me;
	register char *cp, *nextcp;
	register struct mygroup *gr;
	int new;

	mysetgrent(file);
	while ((gr = mygetgrent()) != NULL) {
		if (mastermap) {
			me = getme(&gimap, gr->gr_name, 0, &new);
			if (new)
				giinit(metod(me, struct groupinfo *),
					gr->gr_gid);
			else
				err(0, "\
%s, line %d: duplicate master group %s ignored",
					gr_fname, gr_lineno, gr->gr_name);
		}
		/* assign gid => group mapping */
		me = getme(&groupmap, (char *) NULL, gr->gr_gid, &new);
		*metod(me, char **) = savestr(gr->gr_name);
		/* and set up group members */
		if (!mastermap) {
			for (cp = gr->gr_memlist; cp != NULL; cp = nextcp) {
				nextcp = index(cp, ',');
				if (nextcp != NULL)
					*nextcp++ = 0;
				if (*cp)
					useringroup(cp, gr->gr_name);
			}
		}
	}
	myendgrent();
}

/*
 * Helper function for freegroupmap: free saved group name.
 */
freegroupmapname(me)
	struct mapent *me;
{

	free(*metod(me, char **));
}

/*
 * Toss out the old group map.
 */
freegroupmap()
{

	mapapply(&groupmap, freegroupmapname);
	mapfree(&groupmap);
}

/*
 * Read a group file and set up the current group map.
 */
readgroups(file)
	char *file;
{

	/* throw out the old map, then fill it with the new stuff. */
	freegroupmap();
	dogroupmap(file, 0);
}

/*
 * Read `master' passwd and group files.  These prime the maps with
 * important numbers, e.g., the group `wheel' and the user `root'.
 */
domaster(name)
	char *name;
{
	register struct passwd *pw;
	register struct guinfo *gu;
	struct mapent *me;
	int new;
	char *pwf;

	machine = savestr(strsplice(strsplice("[master: ", name), "]"));
	pwf = savestr(strsplice(name, ".masterpasswd"));
	freegroupmap();
	dogroupmap(strsplice(name, ".mastergroup"), 1);
	mysetpwent(pwf);
	while ((pw = mygetpwent()) != NULL) {
		me = getme(&gumap, pw->pw_name, 0, &new);
		if (new) {
			gu = metod(me, struct guinfo *);
			gu->gu_uid = pw->pw_uid;
			/*
			 * In the presence of multiple master group files,
			 * the call to mastergroup(gidtogroup()) below is
			 * not necessarily an identity function.
			 */
			gu->gu_gid = mastergid(gidtogroup(pw));
		} else
			err(0, "\
%s, line %d: duplicate master user %s ignored",
				pw_fname, pw_lineno, pw->pw_name);
	}
	myendpwent();
	free(pwf);
	free(machine);
}

/*
 * These two unfortunate global variables hold the current
 * output file.  (Output map files are done in a separate module.)
 * (I should move this to a separate module too!)
 */
FILE	*outfp;
char	*outfn;

/*
 * Put a local group map entry out to the uid/gid map file.
 */
putgroupmapent(me)
	struct mapent *me;
{
	int ogid, ngid;

	ogid = me->me_hash;
	ngid = mastergid(*metod(me, char **));
	putmapelt(ogid, ngid, 1);
}

/*
 * Read the passwd and group files for a `slave' machine, and
 * generate a new passwd file.
 *
 * We keep a local user map that maps from UIDs to names, so
 * that two people with the same UID in the existing passwd
 * file wind up with the same UID in the new passwd file.
 */
doslave(host)
	char *host;
{
	register struct passwd *pw;
	register struct guinfo *gu;
	char **namep, *pwf;
	int new, uid;
	struct mapent *me;
	static struct map lumap = { mesize(char *) };

	machine = host;		/* need this globally */

	/*
	 * First read the group file and the passwd file, entering groups
	 * and users.
	 */
	readgroups(strsplice(host, ".group"));
	mysetpwent(pwf = savestr(strsplice(host, ".passwd")));
	while ((pw = mygetpwent()) != NULL) {
		namep = metod(getme(&lumap, (char *) NULL, pw->pw_uid, &new),
			char **);
		if (new) {
			*namep = savestr(pw->pw_name);
			uid = -1;
		} else {
			me = getme(&gumap, *namep, 0, (int *) NULL);
			if (me == NULL)
				err(1, "bug! %s not in gumap", *namep);
			uid = metod(me, struct guinfo *)->gu_uid;
		}
		enteruser(pw, uid);
	}

	/*
	 * Now dump the group map for this host, then dump the passwd
	 * map and the new passwd file.
	 */
	openmapfile(strsplice(host, ".map"));
	mapapply(&groupmap, putgroupmapent);
	mysetpwent(pwf);	/* rewind */
	outfp = wfopen(outfn = savestr(strsplice(host, ".newpw")));
	while ((pw = mygetpwent()) != NULL) {
		gu = metod(getme(&gumap, pw->pw_name, 0, &new),
			struct guinfo *);
		if (new)
			err(1, "bug! new entry for %s@%s", pw->pw_name, host);
		putmapelt(pw->pw_uid, gu->gu_uid, 0);
		fprintf(outfp, "%s:%s:%d:%d:%s:%s:%s\n",
			pw->pw_name, pw->pw_passwd,
			gu->gu_uid, gu->gu_gid,
			pw->pw_gecos, pw->pw_dir, pw->pw_shell);
		if (ferror(outfp))
			err(1, "write(%s): %m", outfn);
	}
	wfclose(outfp, outfn);
	myendpwent();
	closemapfile();
	free(outfn);
	free(pwf);

	/*
	 * Look out!  Borrowing the freeing function from freegroupmap
	 * here... modularity?  What modularity?
	 */
	mapapply(&lumap, freegroupmapname);
	mapfree(&lumap);
}

/*
 * Put a global group info entry to the new group file.
 */
putgrp(me)
	struct mapent *me;
{
	register struct groupinfo *gi = metod(me, struct groupinfo *);
	register char **p;
	register int n;

	(void) fprintf(outfp, "%s:*:%d:", me->me_name, gi->gi_gid);
	p = gi->gi_gu;
	n = gi->gi_nu;
	while (--n >= 0) {
		(void) fputs(*p++, outfp);
		if (n)
			(void) putc(',', outfp);
	}
	(void) putc('\n', outfp);
	if (ferror(outfp))
		err(1, "write(%s): %m", outfn);
}

gengrp()
{

	outfp = wfopen(outfn = "group.new");
	mapapply(&gimap, putgrp);
	wfclose(outfp, outfn);
}
