#ifndef lint
static char rcsid[] = "$Header: zap.c,v 1.5 86/11/05 16:01:22 chris Exp $";
#endif

/*
 * Zap user and group IDs.  Part of the passwd-file-UID-matching
 * program.
 */
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#ifdef sun	/* actually NFS */
#include <sys/vnode.h>
#include <ufs/inode.h>
#include <ufs/fs.h>
#else
#include <sys/inode.h>
#include <sys/fs.h>
#endif
#include <sys/stat.h>
#include <fstab.h>
#include <netinet/in.h>
#include "mf.h"

union {
	struct	fs	sblk;
	char	dummy[SBSIZE];
} un;
#define	sblock	un.sblk

#define	ITABSZ	1024
struct	dinode	itab[ITABSZ];
struct	dinode	*dp;

int	zap;			/* actually do it */
int	fi;			/* descriptor for file system being zapped */
ino_t	ino;			/* current inode */
daddr_t	blk;			/* current inode-bunch-block */
int	dirty;			/* true => current block is dirty */
struct	mapfent *mf;		/* the map */
int	nmf;			/* and its size */

struct	dinode *ginode();
char	*malloc(), *makerawname(), *strcpy(), *strcat();

#define	MAPUID(uid) ((uid) >= nmf ? -1 : mf[uid].mf_uid)
#define	MAPGID(gid) ((gid) >= nmf ? -1 : mf[gid].mf_gid)

main(argc, argv)
	register int argc;
	register char **argv;
{
	register int code = 0, i;

	if (argc <= 1)
		err(1, "usage: %s [ -doit ] fs [ fs ... ]", *argv);
	readmap();
	for (i = 1; i < argc; i++)
		if (i == 1 && strcmp(argv[i], "-doit") == 0)
			zap++;
		else
			code |= doit(argv[i]);
	exit(code);
}

readmap()
{
	register struct mapfent *m;
	int fd;
	short magic;
	struct stat st;
	char fn[100];

	(void) strcpy(fn, "/map");
	fd = open(fn, 0);
	if (fd < 0) {
		fn[0] = '/';
		(void) gethostname(fn + 1, sizeof fn - 6);
		(void) strcat(fn, ".map");
		fd = open(fn, 0);
		if (fd < 0)
			err(1, "cannot open /map or %s: %m", fn);
	}
	if (fstat(fd, &st))
		err(1, "fstat: %m");
	st.st_size -= sizeof (magic);
	if (st.st_size < 0 || (st.st_size % sizeof (*mf)) != 0 ||
	    read(fd, (char *) &magic, sizeof (magic)) != sizeof (magic) ||
	    ntohs(magic) != MAPMAGIC)
		err(1, "%s is not a proper map file", fn);
	if ((mf = (struct mapfent *) malloc((unsigned) st.st_size)) == NULL)
		err(1, "cannot allocate %d map bytes: %m", st.st_size);
	if (read(fd, (char *) mf, st.st_size) != st.st_size)
		err(1, "map read: %m");
	nmf = st.st_size / sizeof (*mf);
	for (m = &mf[nmf]; --m >= mf;) {
		m->mf_uid = ntohl(m->mf_uid);
		m->mf_gid = ntohl(m->mf_gid);
	}
}

doit(fsname)
	char *fsname;
{
	register struct fstab *fs;

	if ((fs = getfsfile(fsname)) == NULL)
		fs = getfsspec(fsname);
	if (fs == NULL)
		return (dodisc(fsname));
	if (strcmp(fs->fs_file, "/") == 0)
		return (dodisc(fs->fs_spec));
	return (dodisc(makerawname(fs->fs_spec)));
}

dodisc(fsdev)
	char *fsdev;
{
	register int cg, i;
	register struct dinode *ip;
	register int newuid, newgid;
	struct stat st;

	fi = open(fsdev, zap ? 2 : 0);
	if (fi < 0) {
		err(0, "cannot open %s: %m", fsdev);
		return (1);
	}
	if (fstat(fi, &st)) {
		err(0, "fstat: %m");
		return (1);
	}
	if ((st.st_mode & S_IFMT) != S_IFCHR &&
	    (st.st_mode & S_IFMT) != S_IFBLK) {
		err(0, "%s is not a character or block device", fsdev);
		return (1);
	}
	(void) printf("%s:\n", fsdev);
	sync();
	sync();
	bread(SBLOCK, (char *)&sblock, SBSIZE);
	ino = -1;
	for (cg = 0; cg < sblock.fs_ncg; cg++) {
		dp = NULL;
		for (i = 0; i < sblock.fs_ipg; i++) {
			ip = ginode();
			if (ip == NULL || ip->di_mode == 0)
				continue;
			newuid = MAPUID(ip->di_uid);
			if (newuid == -1)
				printf("`unowned' file %d (uid %d)\n",
					ino, ip->di_uid);
			newgid = MAPGID(ip->di_gid);
			if (newgid == -1)
				printf("`unowned' file %d (gid %d)\n",
					ino, ip->di_gid);
			if (!zap)
				continue;
			if (newuid == ip->di_uid && newgid == ip->di_gid)
				continue;
			ip->di_uid = newuid;
			ip->di_gid = newgid;
			dirty++;
		}
	}
	flush();
	(void) close(fi);
	sync();
	sync();
	return (0);
}

flush()
{

	if (dirty) {
		bwrite(fsbtodb(&sblock, blk), (char *)itab, sizeof itab);
		dirty = 0;
	}
}

struct dinode *
ginode()
{

	++ino;
	if (dp == NULL || ++dp >= &itab[ITABSZ]) {
		flush();
		blk = itod(&sblock, ino);
		bread(fsbtodb(&sblock, blk), (char *)itab, sizeof itab);
		dp = &itab[ino % INOPB(&sblock)];
	}
	if (ino < ROOTINO)
		return (NULL);
	return (dp);
}

bread(bno, buf, cnt)
	u_long bno;
	char *buf;
{

	(void) lseek(fi, (off_t) dbtob(bno), 0);
	if (read(fi, buf, cnt) != cnt)
		err(1, "read: %m");
}

bwrite(bno, buf, cnt)
	u_long bno;
	char *buf;
{

	(void) lseek(fi, (off_t) dbtob(bno), 0);
	if (write(fi, buf, cnt) != cnt)
		err(1, "write: %m");
}

char *
makerawname(name)
	char *name;
{
	register char *cp;
	char tmp, ch, *rindex();
	static char rawname[MAXPATHLEN];

	(void) strcpy(rawname, name);
	cp = rindex(rawname, '/');
	if (cp == NULL)
		return (name);
	else
		cp++;
	for (ch = 'r'; *cp != '\0'; ) {
		tmp = *cp;
		*cp++ = ch;
		ch = tmp;
	}
	*cp++ = ch;
	*cp = '\0';
	return (rawname);
}
